@@ -38,15 +38,17 @@
 #include "eval.h"
 #include "fit.h"
 #include "gp_hist.h"
+#include "gplocale.h"
+#include "loadpath.h"
 #include "misc.h"
 #include "readline.h"
 #include "setshow.h"
 #include "term_api.h"
 #include "util.h"
-#include "variable.h"
 #include "version.h"
 #include "voxelgrid.h"
 #include "encoding.h"
+#include "xdg.h"
 
 #include <signal.h>
 #include <setjmp.h>
@@ -54,29 +56,17 @@
 #ifdef OS2
 # define INCL_REXXSAA
 # include <os2.h>
-#endif
+#endif /* OS2 */
 
 #if defined(MSDOS) || defined(__EMX__) || (defined(WGP_CONSOLE) && defined(MSVC))
 # include <io.h>
 #endif
 
-#ifdef VMS
-# ifndef __GNUC__
-#  include <unixio.h>
-# endif
-# include <smgdef.h>
-# include <ssdef.h>
-extern int vms_vkid;
-extern smg$create_virtual_keyboard();
-extern int vms_ktid;
-extern smg$create_key_table();
-#endif /* VMS */
-
 #ifdef _WIN32
 # include <windows.h>
 # include "win/winmain.h"
 # include "win/wcommon.h"
-# include <io.h>           // for isatty
+# include <io.h>           /* for isatty */
 #endif /* _WIN32 */
 
 /* GNU readline
@@ -91,14 +81,17 @@ extern smg$create_key_table();
 /* BSD editline
 */
 #ifdef HAVE_LIBEDITLINE
-# include <editline/readline.h>
+#  include <editline/readline.h>
+#  ifdef GNUPLOT_HISTORY
+#    include <histedit.h>
+#  endif
 #endif
 
 /* enable gnuplot history with readline */
 #ifdef GNUPLOT_HISTORY
-# ifndef GNUPLOT_HISTORY_FILE
-#  define GNUPLOT_HISTORY_FILE "~/.gnuplot_history"
-# endif
+#  ifndef GNUPLOT_HISTORY_FILE
+#    define GNUPLOT_HISTORY_FILE "~/.gnuplot_history"
+#  endif
 /*
  * expanded_history_filename points to the value from 'tilde_expand()',
  * which expands '~' to the user's home directory, or $HOME.
@@ -191,56 +184,15 @@ inter(int anint)
 #endif
 }
 
-#ifdef LINUXVGA
-/* utility functions to ensure that setuid gnuplot
- * assumes root privileges only for those parts
- * of the code which require root rights.
+/* Normally we return to the command line at a point equivalent
+ * to program entry, so most context is lost.  All loops and
+ * loaded command files are exited; open files are closed, etc.
  *
- * By "Dr. Werner Fink" <werner@suse.de>
+ * The exception to this is when an error is encountered during a
+ * "fit" command.  In this case the program returns to the same
+ * point it would have reached if the fit command had succeeded,
+ * possibly inside a loop or a loaded command file.
  */
-static uid_t euid, ruid;
-static gid_t egid, rgid;
-static int asked_privi = 0;
-
-void
-drop_privilege()
-{
-    if (!asked_privi) {
-	euid = geteuid();
-	egid = getegid();
-	ruid = getuid();
-	rgid = getgid();
-	asked_privi = 1;
-    }
-    if (setegid(rgid) == -1)
-	(void) fprintf(stderr, "setegid(%d): %s\n",
-		       (int) rgid, strerror(errno));
-    if (seteuid(ruid) == -1)
-	(void) fprintf(stderr, "seteuid(%d): %s\n",
-		       (int) ruid, strerror(errno));
-}
-
-void
-take_privilege()
-{
-    if (!asked_privi) {
-	euid = geteuid();
-	egid = getegid();
-	ruid = getuid();
-	rgid = getgid();
-	asked_privi = 1;
-    }
-    if (setegid(egid) == -1)
-	(void) fprintf(stderr, "setegid(%d): %s\n",
-		       (int) egid, strerror(errno));
-    if (seteuid(euid) == -1)
-	(void) fprintf(stderr, "seteuid(%d): %s\n",
-		       (int) euid, strerror(errno));
-}
-
-#endif /* LINUXVGA */
-
-/* a wrapper for longjmp so we can keep everything local */
 void
 bail_to_command_line()
 {
@@ -248,7 +200,10 @@ bail_to_command_line()
     kill_pending_Pause_dialog();
     ctrlc_flag = FALSE;
 #endif
-    LONGJMP(command_line_env, TRUE);
+    if (fit_env)
+	LONGJMP(*fit_env, TRUE);
+    else
+	LONGJMP(command_line_env, TRUE);
 }
 
 #if defined(_WIN32)
@@ -270,10 +225,6 @@ main(int argc_orig, char **argv)
     static volatile int argc;
     argc = argc_orig;
 
-#ifdef LINUXVGA
-    LINUX_setup();		/* setup VGA before dropping privilege DBT 4/5/99 */
-    drop_privilege();
-#endif
 /* make sure that we really have revoked root access, this might happen if
    gnuplot is compiled without vga support but is installed suid by mistake */
 #ifdef __linux__
@@ -336,7 +287,7 @@ main(int argc_orig, char **argv)
 		*s = NUL;
 	}
     }
-#endif /* DJGPP */
+#endif /* MSDOS || OS2 */
 
 #if (defined(PIPE_IPC) || defined(_WIN32)) && (defined(HAVE_LIBREADLINE) || (defined(HAVE_LIBEDITLINE) && defined(X11)))
     /* Editline needs this to be set before the very first call to readline(). */
@@ -390,9 +341,6 @@ main(int argc_orig, char **argv)
 
 	} else if (!strcmp(argv[i], "-h") || !strcmp(argv[i], "--help")) {
 	    printf( "Usage: gnuplot [OPTION] ... [FILE]\n"
-#ifdef X11
-		    "for X11 options see 'help X11->command-line-options'\n"
-#endif
 		    "  -V, --version\n"
 		    "  -h, --help\n"
 		    "  -p  --persist\n"
@@ -428,15 +376,6 @@ main(int argc_orig, char **argv)
 	}
     }
 
-#ifdef X11
-    /* the X11 terminal removes tokens that it recognizes from argv. */
-    {
-	int n = X11_args(argc, argv);
-	argv += n;
-	argc -= n;
-    }
-#endif
-
     setbuf(stderr, (char *) NULL);
 
 #ifdef HAVE_SETVBUF
@@ -496,24 +435,8 @@ main(int argc_orig, char **argv)
 	}
     }
 
-    /* Need this before show_version is called for the first time */
-
-    if (interactive)
-	show_version(stderr);
-    else
-	show_version(NULL); /* Only load GPVAL_COMPILE_OPTIONS */
-
-    update_gpval_variables(3);  /* update GPVAL_ variables available to user */
-
 #ifdef VMS
-    /* initialise screen management routines for command recall */
-    {
-    unsigned int ierror;
-    if (ierror = smg$create_virtual_keyboard(&vms_vkid) != SS$_NORMAL)
-	done(ierror);
-    if (ierror = smg$create_key_table(&vms_ktid) != SS$_NORMAL)
-	done(ierror);
-    }
+    vms_init_screen();
 #endif /* VMS */
 
     if (!SETJMP(command_line_env, 1)) {
@@ -533,19 +456,45 @@ main(int argc_orig, char **argv)
 #endif
 	init_gadgets();
 
-	/* April 2017: Now that error handling is in place, it is safe parse
-	 * GNUTERM during terminal initialization.
-	 * atexit processing is done in reverse order. We want
-	 * the generic terminal shutdown in term_reset to be executed before
-	 * any terminal specific cleanup requested by individual terminals.
+	/* Now that error handling is in place, it is safe to parse GNUTERM
+	 * during terminal initialization.  However if terminal initialization
+	 * exits via int_error(), the session and history state initialization
+	 * can only be reached via a GOTO from the error return.
 	 */
 	init_terminal();
 	push_terminal(0);	/* remember the initial terminal */
+
+	/* atexit processing is done in reverse order. We want
+	 * the generic terminal shutdown in term_reset to be executed before
+	 * any terminal specific cleanup requested by individual terminals.
+	 */
 	gp_atexit(term_reset);
 
+#ifdef X11
+	/* the X11 terminal removes tokens that it recognizes from argv. */
+	if (term && !strcmp(term->name, "x11")) {
+	    int n = X11_args(argc, argv);
+	    argv += n;
+	    argc -= n;
+	}
+#endif
+	FALLBACK_ONCE_FROM_INIT_FAILURE:
+
+	/* Version 6: defer splash page until after initialization */
+	if (interactive)
+	    show_version(stderr);
+	else
+	    show_version(NULL); /* Only load GPVAL_COMPILE_OPTIONS */
+
+	/* update GPVAL_ variables available to user */
+	update_gpval_variables(3);
+
 	/* Execute commands in ~/.gnuplot */
 	init_session();
 
+	if (interactive)
+	   fprintf(stderr, "\n\tTerminal type is now %s\n", term->name);
+
 	if (interactive && term != 0) {		/* not unknown */
 #ifdef GNUPLOT_HISTORY
 #if (defined(HAVE_LIBREADLINE) || defined(HAVE_LIBEDITLINE)) && !defined(_WIN32)
@@ -554,12 +503,24 @@ main(int argc_orig, char **argv)
 	    expanded_history_filename = gp_strdup(GNUPLOT_HISTORY_FILE);
 	    gp_expand_tilde(&expanded_history_filename);
 #endif
-	    read_history(expanded_history_filename);
+	    if (read_history(expanded_history_filename)) {
+#ifdef USE_XDG_BASEDIR
+		/* use $XDG_STATE_DIR/gnuplot_history only if
+		 * $XDG_STATE_DIR exists or can be created */
+		char *xdgpath = xdg_get_path(kXDGStateHome, "gnuplot_history",
+				FALSE /* no subdir */, TRUE /* create */);
+		if (xdgpath) {
+		    free(expanded_history_filename);
+		    expanded_history_filename = xdgpath;
+		    read_history(expanded_history_filename);
+		}
+#endif /* USE_XDG_BASEDIR */
+	    }
 
 	    /*
 	     * It is safe to ignore the return values of 'atexit()' and
 	     * 'on_exit()'. In the worst case, there is no history of your
-	     * current session and you have to type all again in your next
+	     * current session and you have to type it all again in your next
 	     * session.
 	     */
 	    gp_atexit(wrapper_for_write_history);
@@ -572,9 +533,16 @@ main(int argc_orig, char **argv)
     } else {
 	/* come back here from int_error() */
 	if (!successful_initialization) {
-	    /* Only print the warning once */
+	    /* Only warn and attempt a fall-back on the first failure.
+	     * There is no guarantee that we will come out of this in a
+	     * usable state, but the only truly safe alternative would be
+	     * to give up and exit the program.
+	     */
 	    successful_initialization = TRUE;
-	    fprintf(stderr,"WARNING: Error during initialization\n\n");
+	    fprintf(stderr, "WARNING: Error during initialization\n");
+	    fprintf(stderr, "         Check initialization files and environment variables (e.g. GNUTERM)\n");
+	    change_term("unknown", 7);
+	    goto FALLBACK_ONCE_FROM_INIT_FAILURE;
 	}
 	if (interactive == FALSE)
 	    exit_status = EXIT_FAILURE;
@@ -585,7 +553,15 @@ main(int argc_orig, char **argv)
 	}
 #endif
 
-	load_file_error();	/* if we were in load_file(), cleanup */
+	/* If we were in a "load", "call", or nested bracketed clause 
+	 * clean up the lf_push/pop stack and restore shadowed variables.
+	 */
+	clause_reset_after_error();
+	lf_reset_after_error();
+
+	/* We are certainly no longer in a plot command */
+	inside_plot_command = FALSE;
+
 	SET_CURSOR_ARROW;
 
 #ifdef VMS
@@ -781,11 +757,16 @@ init_session()
 	/* Undefine any previously-used variables */
 	del_udv_by_name("",TRUE);
 
+	/* Clear any previous customization of linetypes */
+	while (first_perm_linestyle != NULL)
+	    delete_linestyle(&first_perm_linestyle, NULL, first_perm_linestyle);
+
 	/* Restore default colors before loading local preferences */
 	set_colorsequence(1);
 
 	/* Reset program variables not handled by 'reset' */
 	overflow_handling = INT64_OVERFLOW_TO_FLOAT;
+	suppress_warnings = FALSE;
 
 	/* Reset voxel data structures if supported */
 	init_voxelsupport();
@@ -795,19 +776,19 @@ init_session()
 	 */
 	reset_command();	/* FIXME: this does c_token++ */
 	load_rcfile(0);		/* System-wide gnuplotrc if configured */
-	load_rcfile(1);		/* ./.gnuplot if configured */
 
 	/* After this point we allow pipes and system commands */
 	successful_initialization = TRUE;
 
 	load_rcfile(2);		/* ~/.gnuplot */
+	load_rcfile(3);		/* ~/.config/gnuplot/gnuplotrc */
 }
 
 /*
  * Read commands from an initialization file.
  * where = 0: look for gnuplotrc in system shared directory
- * where = 1: look for .gnuplot in current directory
- * where = 2: look for .gnuplot in home directory
+ * where = 1: look for .gnuplot in current directory (DEPRECATED)
+ * where = 2: look for .gnuplot in user's home directory
  */
 static void
 load_rcfile(int where)
@@ -830,14 +811,6 @@ load_rcfile(int where)
 	plotrc = fopen(rcfile, "r");
 #endif
 
-    } else if (where == 1) {
-#ifdef USE_CWDRC
-    /* Allow check for a .gnuplot init file in the current directory */
-    /* This is a security risk, as someone might leave a malicious   */
-    /* init file in a shared directory.                              */
-	plotrc = fopen(PLOTRC, "r");
-#endif /* !USE_CWDRC */
-
     } else if (where == 2 && user_homedir) {
 	/* length of homedir + directory separator + length of file name + \0 */
 	int len = (user_homedir ? strlen(user_homedir) : 0) + 1 + strlen(PLOTRC) + 1;
@@ -845,6 +818,13 @@ load_rcfile(int where)
 	strcpy(rcfile, user_homedir);
 	PATH_CONCAT(rcfile, PLOTRC);
 	plotrc = fopen(rcfile, "r");
+    } else if (where == 3) {
+#ifdef USE_XDG_BASEDIR
+	rcfile = xdg_get_path(kXDGConfigHome, "gnuplotrc",
+			TRUE /* use subdir */, FALSE /* not create */);
+	if (rcfile)
+	    plotrc = fopen(rcfile, "r");
+#endif
     }
 
     if (plotrc) {
